@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2017  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2017-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -140,6 +140,20 @@ static void interface_update_properties(struct proxy_interface *proxy,
 		proxy_interface_property_update(proxy, name, NULL);
 }
 
+static char *strdup_quoted_for_spaces(const char *str)
+{
+	const char *p;
+
+	for (p = str; *p; p++) {
+		if (*p != ' ')
+			continue;
+
+		return l_strdup_printf("\"%s\"", str);
+	}
+
+	return l_strdup(str);
+}
+
 char *proxy_property_str_completion(const struct proxy_interface_type *type,
 					proxy_property_match_func_t function,
 					const char *property_name,
@@ -177,7 +191,7 @@ char *proxy_property_str_completion(const struct proxy_interface_type *type,
 		if (!str)
 			goto done;
 
-		return l_strdup(str);
+		return strdup_quoted_for_spaces(str);
 	}
 done:
 	l_queue_destroy(match, NULL);
@@ -285,6 +299,7 @@ struct proxy_callback_data {
 static void proxy_callback(struct l_dbus_message *message, void *user_data)
 {
 	struct proxy_callback_data *callback_data = user_data;
+	const struct proxy_interface *proxy;
 	const char *name;
 	const char *text;
 
@@ -294,9 +309,16 @@ static void proxy_callback(struct l_dbus_message *message, void *user_data)
 	if (command_is_interactive_mode())
 		return;
 
-	if (l_dbus_message_get_error(message, &name, &text))
+	if (l_dbus_message_get_error(message, &name, &text)) {
 		command_set_exit_status(EXIT_FAILURE);
+		goto quit;
+	}
 
+	proxy = callback_data->user_data;
+	if (!strcmp(proxy->type->interface, IWD_AGENT_MANAGER_INTERFACE))
+		return;
+
+quit:
 	l_main_quit();
 }
 
@@ -682,14 +704,18 @@ static void get_managed_objects_callback(struct l_dbus_message *message,
 	while (l_dbus_message_iter_next_entry(&objects, &path, &object))
 		proxy_interface_create(path, &object);
 
-	if (!command_is_interactive_mode()) {
-		command_noninteractive_trigger();
+	if (!command_is_interactive_mode() && !command_has_options())
+		goto proceed;
+
+	if (!agent_manager_register_agent()) {
+		l_main_quit();
 
 		return;
 	}
 
-	if (!agent_manager_register_agent()) {
-		l_main_quit();
+proceed:
+	if (!command_is_interactive_mode()) {
+		command_noninteractive_trigger();
 
 		return;
 	}
@@ -734,6 +760,8 @@ static void service_disappeared_callback(struct l_dbus *dbus,
 
 	l_queue_clear(proxy_interfaces, proxy_interface_destroy);
 
+	command_reset_default_entities();
+
 	display_disable_cmd_prompt();
 }
 
@@ -800,7 +828,9 @@ bool dbus_proxy_exit(void)
 {
 	struct interface_type_desc *desc;
 
-	if (command_is_interactive_mode())
+	if (command_is_interactive_mode() ||
+			(!command_is_interactive_mode() &&
+							command_has_options()))
 		agent_manager_unregister_agent();
 
 	for (desc = __start___interface; desc < __stop___interface; desc++) {
