@@ -20,6 +20,10 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <ctype.h>
 #include <stdio.h>
@@ -31,7 +35,6 @@
 
 #include "src/dbus.h"
 #include "src/simauth.h"
-#include "src/backtrace.h"
 
 /*
  * This plugin takes care of all the communication with ofono in order to
@@ -90,7 +93,7 @@ struct user_cb {
 static uint32_t ofono_watch;
 static uint32_t modem_add_watch;
 static uint32_t modem_removed_watch;
-struct l_queue *modems;
+static struct l_queue *modems;
 
 static struct user_cb *new_cb(void *func, void *data, bool is_gsm)
 {
